@@ -12,7 +12,7 @@
 #endif
 
 //Description: update B and F.
-void updateBF(double *B, double *F, double *c, double *C, double *D, double *d, int *nnIndxLU, int *CIndx, int n, double sigmaSq, double phi, double nu, int covModel, double *bk, double nuUnifb){
+void updateBF(double *B, double *F, double *c, double *C, double *coords, int *nnIndx, int *nnIndxLU, int n, int m, double sigmaSq, double phi, double nu, int covModel, double *bk, double nuUnifb){
     
   int i, k, l;
   int info = 0;
@@ -24,9 +24,11 @@ void updateBF(double *B, double *F, double *c, double *C, double *D, double *d,
   //bk must be 1+(int)floor(alpha) * nthread
   int nb = 1+static_cast<int>(floor(nuUnifb));
   int threadID = 0;
+  double e;
+  int mm = m*m;
   
 #ifdef _OPENMP
-#pragma omp parallel for private(k, l, info, threadID)
+#pragma omp parallel for private(k, l, info, threadID, e)
 #endif
     for(i = 0; i < n; i++){
 #ifdef _OPENMP
@@ -34,15 +36,17 @@ void updateBF(double *B, double *F, double *c, double *C, double *D, double *d,
 #endif
       if(i > 0){
 	for(k = 0; k < nnIndxLU[n+i]; k++){
-	  c[nnIndxLU[i]+k] = sigmaSq*spCor(d[nnIndxLU[i]+k], phi, nu, covModel, &bk[threadID*nb]);
+	  e = dist2(coords[i], coords[n+i], coords[nnIndx[nnIndxLU[i]+k]], coords[n+nnIndx[nnIndxLU[i]+k]]);
+	  c[m*threadID+k] = sigmaSq*spCor(e, phi, nu, covModel, &bk[threadID*nb]);
 	  for(l = 0; l <= k; l++){
-	    C[CIndx[i]+l*nnIndxLU[n+i]+k] = sigmaSq*spCor(D[CIndx[i]+l*nnIndxLU[n+i]+k], phi, nu, covModel, &bk[threadID*nb]); 
+	    e = dist2(coords[nnIndx[nnIndxLU[i]+k]], coords[n+nnIndx[nnIndxLU[i]+k]], coords[nnIndx[nnIndxLU[i]+l]], coords[n+nnIndx[nnIndxLU[i]+l]]); 
+	    C[mm*threadID+l*nnIndxLU[n+i]+k] = sigmaSq*spCor(e, phi, nu, covModel, &bk[threadID*nb]); 
 	  }
 	}
-	F77_NAME(dpotrf)(&lower, &nnIndxLU[n+i], &C[CIndx[i]], &nnIndxLU[n+i], &info); if(info != 0){error("c++ error: dpotrf failed\n");}
-	F77_NAME(dpotri)(&lower, &nnIndxLU[n+i], &C[CIndx[i]], &nnIndxLU[n+i], &info); if(info != 0){error("c++ error: dpotri failed\n");}
-	F77_NAME(dsymv)(&lower, &nnIndxLU[n+i], &one, &C[CIndx[i]], &nnIndxLU[n+i], &c[nnIndxLU[i]], &inc, &zero, &B[nnIndxLU[i]], &inc);
-	F[i] = sigmaSq - F77_NAME(ddot)(&nnIndxLU[n+i], &B[nnIndxLU[i]], &inc, &c[nnIndxLU[i]], &inc);
+	F77_NAME(dpotrf)(&lower, &nnIndxLU[n+i], &C[mm*threadID], &nnIndxLU[n+i], &info); if(info != 0){error("c++ error: dpotrf failed\n");}
+	F77_NAME(dpotri)(&lower, &nnIndxLU[n+i], &C[mm*threadID], &nnIndxLU[n+i], &info); if(info != 0){error("c++ error: dpotri failed\n");}
+	F77_NAME(dsymv)(&lower, &nnIndxLU[n+i], &one, &C[mm*threadID], &nnIndxLU[n+i], &c[m*threadID], &inc, &zero, &B[nnIndxLU[i]], &inc);
+	F[i] = sigmaSq - F77_NAME(ddot)(&nnIndxLU[n+i], &B[nnIndxLU[i]], &inc, &c[m*threadID], &inc);
       }else{
 	B[i] = 0;
 	F[i] = sigmaSq;
@@ -53,11 +57,11 @@ void updateBF(double *B, double *F, double *c, double *C, double *D, double *d,
 
 extern "C" {
   
-  SEXP sNNGP(SEXP y_r, SEXP X_r, SEXP p_r, SEXP n_r, SEXP m_r, SEXP coords_r, SEXP covModel_r, 
+  SEXP sNNGP(SEXP y_r, SEXP X_r, SEXP p_r, SEXP n_r, SEXP m_r, SEXP coords_r, SEXP covModel_r, SEXP nnIndx_r, SEXP nnIndxLU_r,
 	     SEXP sigmaSqIG_r, SEXP tauSqIG_r, SEXP phiUnif_r, SEXP nuUnif_r, 
 	     SEXP betaStarting_r, SEXP sigmaSqStarting_r, SEXP tauSqStarting_r, SEXP phiStarting_r, SEXP nuStarting_r,
 	     SEXP sigmaSqTuning_r, SEXP tauSqTuning_r, SEXP phiTuning_r, SEXP nuTuning_r, 
-	     SEXP nSamples_r, SEXP sType_r, SEXP returnNNIndx_r, SEXP nThreads_r, SEXP verbose_r, SEXP nReport_r){
+	     SEXP nSamples_r, SEXP nThreads_r, SEXP verbose_r, SEXP nReport_r){
     
     int h, i, j, k, l, s, info, nProtect=0;
     const int inc = 1;
@@ -78,7 +82,8 @@ extern "C" {
     int n = INTEGER(n_r)[0];
     int m = INTEGER(m_r)[0];
     double *coords = REAL(coords_r);
-    
+    int *nnIndx = INTEGER(nnIndx_r);
+    int *nnIndxLU = INTEGER(nnIndxLU_r);
     int covModel = INTEGER(covModel_r)[0];
     std::string corName = getCorName(covModel);
         
@@ -165,29 +170,8 @@ extern "C" {
       tuning[nuIndx] = REAL(nuTuning_r)[0];
     }
 
-    //allocated for the nearest neighbor index vector (note, first location has no neighbors).
-    int nIndx = static_cast<int>(static_cast<double>(1+m)/2*m+(n-m-1)*m);
-    //int *nnIndx = (int *) R_alloc(nIndx, sizeof(int));  
-    SEXP nnIndx_r; PROTECT(nnIndx_r = allocVector(INTSXP, nIndx)); nProtect++; int *nnIndx = INTEGER(nnIndx_r);
-    double *d = (double *) R_alloc(nIndx, sizeof(double));
-    int *nnIndxLU = (int *) R_alloc(2*n, sizeof(int)); //first column holds the nnIndx index for the i-th location and the second columns holds the number of neighbors the i-th location has (the second column is a bit of a waste but will simplifying some parallelization).
-
-    //make the neighbor index
-    if(verbose){
-      Rprintf("----------------------------------------\n");
-      Rprintf("\tBuilding neighbor index\n");
-      #ifdef Win32
-        R_FlushConsole();
-      #endif
-    }
-    
-    if(INTEGER(sType_r)[0] == 0){
-      mkNNIndx(n, m, coords, nnIndx, d, nnIndxLU);
-    }else{
-      mkNNIndxTree0(n, m, coords, nnIndx, d, nnIndxLU);
-    }
-
     //allocate for the U index vector that keep track of which locations have the i-th location as a neighbor
+    int nIndx = static_cast<int>(static_cast<double>(1+m)/2*m+(n-m-1)*m);
     int *uIndx = (int *) R_alloc(nIndx, sizeof(int)); //U indexes 
     
     //first column holds the uIndx index for i-th location and second column holds
@@ -215,34 +199,13 @@ extern "C" {
     }
  
     //other stuff
+    int mm = m*m;
     double *B = (double *) R_alloc(nIndx, sizeof(double));
     double *F = (double *) R_alloc(n, sizeof(double));
     double *BCand = (double *) R_alloc(nIndx, sizeof(double));
     double *FCand = (double *) R_alloc(n, sizeof(double));
-    double *c =(double *) R_alloc(nIndx, sizeof(double));
-    
-    int *CIndx = (int *) R_alloc(2*n, sizeof(int)); //index for D and C.
-    for(i = 0, j = 0; i < n; i++){//zero should never be accessed
-      j += nnIndxLU[n+i]*nnIndxLU[n+i];
-      if(i == 0){
-	CIndx[n+i] = 0;
-	CIndx[i] = 0;
-      }else{
-	CIndx[n+i] = nnIndxLU[n+i]*nnIndxLU[n+i]; 
-	CIndx[i] = CIndx[n+i-1] + CIndx[i-1];
-      }
-    }
-    
-    double *C = (double *) R_alloc(j, sizeof(double)); zeros(C, j);
-    double *D = (double *) R_alloc(j, sizeof(double)); zeros(D, j);
-    
-    for(i = 0; i < n; i++){
-      for(k = 0; k < nnIndxLU[n+i]; k++){   
-	for(l = 0; l <= k; l++){
-	  D[CIndx[i]+l*nnIndxLU[n+i]+k] = dist2(coords[nnIndx[nnIndxLU[i]+k]], coords[n+nnIndx[nnIndxLU[i]+k]], coords[nnIndx[nnIndxLU[i]+l]], coords[n+nnIndx[nnIndxLU[i]+l]]);
-	}
-      }
-    }
+    double *c =(double *) R_alloc(m*nThreads, sizeof(double));
+    double *C = (double *) R_alloc(mm*nThreads, sizeof(double));
 
     //return stuff  
     SEXP betaSamples_r, thetaSamples_r, wSamples_r;
@@ -251,7 +214,8 @@ extern "C" {
     PROTECT(wSamples_r = allocMatrix(REALSXP, n, nSamples)); nProtect++;
     
     //other stuff
-    double logPostCand, logPostCurrent, logDet, accept = 0, batchAccept = 0, status = 0;
+    double logPostCand, logPostCurrent, logDet;
+    int accept = 0, batchAccept = 0, status = 0;
     int jj, kk, pp = p*p;
     double *tmp_pp = (double *) R_alloc(pp, sizeof(double));
     double *tmp_p = (double *) R_alloc(p, sizeof(double));
@@ -275,8 +239,8 @@ extern "C" {
     }
 
     if(corName == "matern"){nu = theta[nuIndx];}
-    updateBF(B, F, c, C, D, d, nnIndxLU, CIndx, n, theta[sigmaSqIndx], theta[phiIndx], nu, covModel, bk, nuUnifb);
-    
+    updateBF(B, F, c, C, coords, nnIndx, nnIndxLU, n, m, theta[sigmaSqIndx], theta[phiIndx], nu, covModel, bk, nuUnifb);
+      
     GetRNGstate();
     
     for(s = 0; s < nSamples; s++){
@@ -369,7 +333,7 @@ extern "C" {
       ///////////////
       //current
       if(corName == "matern"){nu = theta[nuIndx];}
-      updateBF(B, F, c, C, D, d, nnIndxLU, CIndx, n, theta[sigmaSqIndx], theta[phiIndx], nu, covModel, bk, nuUnifb);
+      updateBF(B, F, c, C, coords, nnIndx, nnIndxLU, n, m, theta[sigmaSqIndx], theta[phiIndx], nu, covModel, bk, nuUnifb);
       
       a = 0;
       logDet = 0;
@@ -404,8 +368,8 @@ extern "C" {
       	nuCand = logitInv(rnorm(logit(theta[nuIndx], nuUnifa, nuUnifb), tuning[nuIndx]), nuUnifa, nuUnifb);
       }
       
-      updateBF(BCand, FCand, c, C, D, d, nnIndxLU, CIndx, n, theta[sigmaSqIndx], phiCand, nuCand, covModel, bk, nuUnifb);
-      
+      updateBF(BCand, FCand, c, C, coords, nnIndx, nnIndxLU, n, m, theta[sigmaSqIndx], phiCand, nuCand, covModel, bk, nuUnifb);
+            
       a = 0;
       logDet = 0;
       
@@ -476,10 +440,6 @@ extern "C" {
     //make return object
     SEXP result_r, resultName_r;
     int nResultListObjs = 3;
-
-    if(INTEGER(returnNNIndx_r)[0]){
-      nResultListObjs++;
-    }
     
     PROTECT(result_r = allocVector(VECSXP, nResultListObjs)); nProtect++;
     PROTECT(resultName_r = allocVector(VECSXP, nResultListObjs)); nProtect++;
@@ -492,11 +452,6 @@ extern "C" {
 
     SET_VECTOR_ELT(result_r, 2, wSamples_r);
     SET_VECTOR_ELT(resultName_r, 2, mkChar("p.w.samples"));
-
-    if(INTEGER(returnNNIndx_r)[0]){
-      SET_VECTOR_ELT(result_r, 3, nnIndx_r);
-      SET_VECTOR_ELT(resultName_r, 3, mkChar("n.indx")); 
-    }
 	
     namesgets(result_r, resultName_r);
     
